"""
Unit conversion for ebas input and output
"""

import re
import atmos_phys.convert.gas as gas
import atmos_phys.convert.aerosol as aerosol
import atmos_phys.convert.precipitation as precipitation
import atmos_phys.convert.electric as electric
from atmos_phys.convert.temperature import ConvertTemperature
from atmos_phys.convert.pressure import ConvertPressure
from ebas.domain.masterdata.pm import EbasMasterPM


class NoConversion(Exception):
    """
    Exception raised when no conversion is possible.
    """
    def __init__(self, msgs):
        """
        Init the object
        """
        self.msgs = msgs

class UnitConvert(object):
    """
    Unit conversion for ebas input and output
    """

    # Configuration for EBAS unit conversion on import/export
    # regime, matrix, comp_name (and ebas_unit on export) identify the parameter
    # import: import conversion should be done
    #     units: accepted units for conversion on import
    #     need_stdcond: True: only converted if std conditions are given
    #                         (no conversion if not given in metadata)
    #                   False: not needed, conversion is possible anyway
    #     use_stdcond: standard conditions to be used for import conversion
    #                  (pressuer hPa, temp K) or (None, None) (no std cond used)
    #                  if need_stdcond: this is overruled
    #     set_stdcond: std condistions will be set to after conversion
    #                  (pressuer hPa, temp K)
    #                  (None, None) set to none
    #                  None: do not change
    #     roundoffset: round offset used for import conversion
    # export: export conversion should be done
    #     unit: unit to be converted on export
    #     need_stdcond: True: only converted if std conditions are given
    #                         (no conversion if not given in metadata)
    #                   False: not needed, conversion is possible anyway
    #     use_stdcond: standard conditions to be used for export conversion
    #                  (pressuer hPa, temp K) or (None, None) (no std cond used)
    #                  if need_stdcond: this is overruled
    #     set_stdcond: std condistions will be set to after export conversion
    #                  (pressuer hPa, temp K)
    #                  (None, None) set to none
    #                  None: do not change
    #     roundoffset: round offset used for export conversion
    #     maxround: maxround used for export conversion
    # cvt_class: conversion class to be used
    CONVERSIONS = [
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'carbon_monoxide',
            "ebas_unit": 'nmol/mol',
            "import": {
                "units": ['mg/m3', 'ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            # no export conversion
            "cvt_class": gas.ConvertCarbonMonoxide
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "nitrogen_monoxide",
            "ebas_unit": "ug N/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertNOx
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "nitrogen_dioxide",
            "ebas_unit": "ug N/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertNOx
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "NOx",
            "ebas_unit": "ug N/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertNOx
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "ammonia",
            "ebas_unit": "ug N/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertAmmonia
        },
        {   # additional option for import conversion massconc -> massconc
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'ammonia',
            "ebas_unit": 'ug N/m3',
            "import": {
                "units": ['ug/m3'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 0,  # no roundoffset, will not be re-exported
            },
            "cvt_class": gas.ConvertAmmonia
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "sulphur_dioxide",
            "ebas_unit": "ug S/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertSulphurDioxide
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "hydrochloric_acid",
            "ebas_unit": "ug Cl/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertHydrochloricAcid
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "nitric_acid",
            "ebas_unit": "ug N/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertNitricAcid
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "nitrous_acid",
            "ebas_unit": "ug N/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertNitrousAcid
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "ozone",
            "ebas_unit": "ug/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb'],
                "need_stdcond": False,
                # ozone: exceptional std cond.: 293.15 K
                "use_stdcond": (1013.25, 293.15),
                "set_stdcond": (1013.25, 293.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertOzone
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "ethanal",
            "ebas_unit": "ug/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertEthanal
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "ethanol",
            "ebas_unit": "ug/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertEthanol
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "hexanal",
            "ebas_unit": "ug/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "pmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertHexanal
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "n-butanal",
            "ebas_unit": "ug/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "pmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertNButanal
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "naphthalene",
            "ebas_unit": "ng/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                # special case convert from ug/m3 to ng/m3 SEE BELOW
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "pmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertNaphthalene
        },
        {   # additional option for import conversion massconc -> massconc
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'naphthalene',
            "ebas_unit": 'ng/m3',
            "import": {
                "units": ['ug/m3'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 0,  # no roundoffset needed, factor 1000.0
            },
            "cvt_class": gas.ConvertNaphthalene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "pentanal",
            "ebas_unit": "ug/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "pmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertPentanal
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "2-methylpropenal",
            "ebas_unit": "ug/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "pmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.Convert2Methylpropenal
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "benzaldehyde",
            "ebas_unit": "ug/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "pmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertBenzaldehyde
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "2-propanol",
            "ebas_unit": "ug/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "pmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.Convert2Propanol
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "methanal",
            "ebas_unit": "ug/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertMethanal
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "propanone",
            "ebas_unit": "ug/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertPropanone
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "propanal",
            "ebas_unit": "ug/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertPropanal
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "n-propanol",
            "ebas_unit": "ug/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertNPropanol
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "ethanedial",
            "ebas_unit": "ug/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertEthanedial
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "2-oxopropanal",
            "ebas_unit": "ug/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.Convert2Oxopropanal
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "2-propenal",
            "ebas_unit": "ug/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "pmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.Convert2Propenal
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "3-buten-2-one",
            "ebas_unit": "ug/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "pmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.Convert3Buten2One
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "butanone",
            "ebas_unit": "ug/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "pmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertButanone
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": "styrene",
            "ebas_unit": "ug/m3",
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "pmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertStyrene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": '1-2-3-trimethylbenzene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ug/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.Convert123Trimethylbenzene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": '1-2-4-trimethylbenzene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ug/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.Convert124Trimethylbenzene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": '1-3-5-trimethylbenzene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ug/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.Convert135Trimethylbenzene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": '1-ethyl-3-methylbenzene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ug/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.Convert1Ethyl3Methylbenzene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": '1-ethyl-4-methylbenzene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ug/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.Convert1Ethyl4Methylbenzene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": '3-carene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ug/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.Convert3Carene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'acenaphthene',
            "ebas_unit": 'ng/m3',
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "pmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertAcenaphthene
        },
        {   # additional option for import conversion massconc -> massconc
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'acenaphthene',
            "ebas_unit": 'ng/m3',
            "import": {
                "units": ['ug/m3'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 0,  # no roundoffset needed, factor 1000.0
            },
            "cvt_class": gas.ConvertAcenaphthene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'acenaphthylene',
            "ebas_unit": 'ng/m3',
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "pmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertAcenaphthylene
        },
        {   # additional option for import conversion massconc -> massconc
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'acenaphthylene',
            "ebas_unit": 'ng/m3',
            "import": {
                "units": ['ug/m3'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 0,  # no roundoffset needed, factor 1000.0
            },
            "cvt_class": gas.ConvertAcenaphthylene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'anthracene',
            "ebas_unit": 'ng/m3',
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "pmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertAnthracene
        },
        {   # additional option for import conversion massconc -> massconc
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'anthracene',
            "ebas_unit": 'ng/m3',
            "import": {
                "units": ['ug/m3'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 0,  # no roundoffset needed, factor 1000.0
            },
            "cvt_class": gas.ConvertAnthracene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'fluorene',
            "ebas_unit": 'ng/m3',
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "pmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertFluorene
        },
        {   # additional option for import conversion massconc -> massconc
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'fluorene',
            "ebas_unit": 'ng/m3',
            "import": {
                "units": ['ug/m3'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 0,  # no roundoffset needed, factor 1000.0
            },
            "cvt_class": gas.ConvertFluorene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'alpha-pinene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertAlphaPinene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'benzene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertBenzene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'beta-pinene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertBetaPinene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'camphene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertCamphene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'ethylbenzene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertEthylbenzene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'limonene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertLimonene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'linalool',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertLinalool
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'm-p-xylene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertMPXylene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'myrcene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertMyrcene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'n-decane',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertNDecane
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'n-dodecane',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertNDodecane
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'n-nonane',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertNNonane
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'n-octane',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertNOctane
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'n-pentadecane',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertNPentadecane
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'n-propylbenzene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertNPropylbenzene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'n-tetradecane',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertNTetradecane
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'n-tridecane',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertNTridecane
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'n-undecane',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertNUndecane
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'o-xylene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertOXylene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'p-cymene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertPCymene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'tert-butylbenzene',
            "ebas_unit": 'ug/m3',
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                # special case convert from ng/m3 to ug/m3 SEE BELOW
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "pmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertTertButylbenzene,
        },
        {   # additional option for import conversion massconc -> massconc
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'tert-butylbenzene',
            "ebas_unit": 'ug/m3',
            "import": {
                "units": ['ng/m3'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 0,  # no roundoffset needed, factor 1000.0
            },
            "cvt_class": gas.ConvertTertButylbenzene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": '1-ethyl-2-methylbenzene',
            "ebas_unit": 'ug/m3',
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                # special case convert from ng/m3 to ug/m3 SEE BELOW
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "pmol/mol",
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.Convert1Ethyl2Methylbenzene,
        },
        {   # additional option for import conversion massconc -> massconc
            "regime": "IMG",
            "matrix": "air",
            "comp_name": '1-ethyl-2-methylbenzene',
            "ebas_unit": 'ug/m3',
            "import": {
                "units": ['ng/m3'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 0,  # no roundoffset needed, factor 1000.0
            },
            "cvt_class": gas.Convert1Ethyl2Methylbenzene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": '1-butene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.Convert1Butene,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'chloroethene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertChloroethene,
        },

        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'methane',
            "ebas_unit": 'nmol/mol',
            "import": {
                "units": ['mg/m3', 'ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "mg/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertMethane,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'ethane',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertEthane,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'propane',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertPropane,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": '2-methylpropane',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.Convert2Methylpropane,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'n-hexadecane',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertNHexadecane,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'n-butane',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertNButane,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": '1-3-butadiene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.Convert13Butadiene,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'trans-2-butene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertTrans2Butene,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'cis-2-butene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertCis2Butene,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": '2-methylbutane',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.Convert2Methylbutane,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": '1-pentene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.Convert1Pentene,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'n-pentane',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertNPentane,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'trans-2-pentene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertTrans2Pentene,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'isoprene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertIsoprene,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'cis-2-pentene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertCis2Pentene,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": '2-methyl-2-butene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.Convert2Methyl2Butene,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'dichloromethane',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertDichloromethane,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": '2-methylpentane',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.Convert2Methylpentane,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": '3-methylpentane',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.Convert3Methylpentane,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": '1-hexene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.Convert1Hexene,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'n-hexane',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertNHexane,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'trichloroethane',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertTrichloroethane,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": '1-2-dichloroethane',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.Convert12Dichloroethane,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": '2-2-4-trimethylpentane',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.Convert224Trimethylpentane,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'n-heptane',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertNHeptane,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'trichloroethene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertTrichloroethene,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'tetrachloroethene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertTetrachloroethene,
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'sabinene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertSabinene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'terpinolene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertTerpinolene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'toluene',
            "ebas_unit": 'pmol/mol',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": True,
                "use_stdcond": (None, None),
                "set_stdcond": (None, None),
                "roundoffset": 1,
            },
            "export": {
                "unit": "ng/m3",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertToluene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": '1-4-dichlorobenzene',
            "ebas_unit": 'pg/m3',
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.Convert14Dichlorobenzene
        },
        {   # additional option for import conversion massconc -> massconc
            "regime": "IMG",
            "matrix": "air",
            "comp_name": '1-4-dichlorobenzene',
            "ebas_unit": 'pg/m3',
            "import": {
                "units": ['ug/m3', 'ng/m3'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 0,  # no roundoffset needed, factor 1000.0
            },
            # no export conversion
            "cvt_class": gas.Convert14Dichlorobenzene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'eucalyptol',
            "ebas_unit": 'ug/m3',
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertEucalyptol
        },
        {   # additional option for import conversion massconc -> massconc
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'eucalyptol',
            "ebas_unit": 'ug/m3',
            "import": {
                "units": ['ng/m3', 'pg/m3'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 0,  # no roundoffset needed, factor 1000.0
            },
            # no export conversion
            "cvt_class": gas.ConvertEucalyptol
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'longicyclene',
            "ebas_unit": 'ug/m3',
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertLongicyclene
        },
        {   # additional option for import conversion massconc -> massconc
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'longicyclene',
            "ebas_unit": 'ug/m3',
            "import": {
                "units": ['ng/m3', 'pg/m3'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 0,  # no roundoffset needed, factor 1000.0
            },
            # no export conversion
            "cvt_class": gas.ConvertLongicyclene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'iso-longifolene',
            "ebas_unit": 'ug/m3',
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertIsoLongifolene
        },
        {   # additional option for import conversion massconc -> massconc
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'iso-longifolene',
            "ebas_unit": 'ug/m3',
            "import": {
                "units": ['ng/m3', 'pg/m3'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 0,  # no roundoffset needed, factor 1000.0
            },
            # no export conversion
            "cvt_class": gas.ConvertIsoLongifolene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'beta-caryophyllene',
            "ebas_unit": 'ug/m3',
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertBetaCaryophyllene
        },
        {   # additional option for import conversion massconc -> massconc
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'beta-caryophyllene',
            "ebas_unit": 'ug/m3',
            "import": {
                "units": ['ng/m3', 'pg/m3'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 0,  # no roundoffset needed, factor 1000.0
            },
            # no export conversion
            "cvt_class": gas.ConvertBetaCaryophyllene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'beta-farnesene',
            "ebas_unit": 'ug/m3',
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertBetaFarnesene
        },
        {   # additional option for import conversion massconc -> massconc
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'beta-farnesene',
            "ebas_unit": 'ug/m3',
            "import": {
                "units": ['ng/m3', 'pg/m3'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 0,  # no roundoffset needed, factor 1000.0
            },
            # no export conversion
            "cvt_class": gas.ConvertBetaFarnesene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'alpha-humulene',
            "ebas_unit": 'ug/m3',
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertAlphaHumulene
        },
        {   # additional option for import conversion massconc -> massconc
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'alpha-humulene',
            "ebas_unit": 'ug/m3',
            "import": {
                "units": ['ng/m3', 'pg/m3'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 0,  # no roundoffset needed, factor 1000.0
            },
            # no export conversion
            "cvt_class": gas.ConvertAlphaHumulene
        },
        {
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'nopinone',
            "ebas_unit": 'ug/m3',
            "import": {
                "units": ['nmol/mol', 'ppbv', 'ppb', 'pmol/mol', 'pptv', 'ppt'],
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": 1,
            },
            "export": {
                "unit": "nmol/mol",
                "need_stdcond": False,
                "use_stdcond": (1013.25, 273.15),
                "set_stdcond": (1013.25, 273.15),
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": gas.ConvertNopinone
        },
        {   # additional option for import conversion massconc -> massconc
            "regime": "IMG",
            "matrix": "air",
            "comp_name": 'nopinone',
            "ebas_unit": 'ug/m3',
            "import": {
                "units": ['ng/m3', 'pg/m3'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 0,  # no roundoffset needed, factor 1000.0
            },
            # no export conversion
            "cvt_class": gas.ConvertNopinone
        },


        #
        # Aerosols
        #
        {
            "regime": "IMG",
            "matrix": "ALL_AEROSOL",
            "comp_name": "nitrate",
            "ebas_unit": "ug N/m3",
            "import": {
                "units": ['ug/m3'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 1,
            },
            "export": {
                "unit": "ug/m3",
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": aerosol.ConvertNitrate
        },

        {
            "regime": "IMG",
            "matrix": "ALL_AEROSOL",
            "comp_name": "nitrite",
            "ebas_unit": "ug N/m3",
            "import": {
                "units": ['ug/m3'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 1,
            },
            "export": {
                "unit": "ug/m3",
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": aerosol.ConvertNitrite
        },

        {
            "regime": "IMG",
            "matrix": "ALL_AEROSOL",
            "comp_name": "ammonium",
            "ebas_unit": "ug N/m3",
            "import": {
                "units": ['ug/m3'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 1,
            },
            "export": {
                "unit": "ug/m3",
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": aerosol.ConvertAmmonium
        },

        {
            "regime": "IMG",
            "matrix": "ALL_AEROSOL",
            "comp_name": "sulphate_total",
            "ebas_unit": "ug S/m3",
            "import": {
                "units": ['ug/m3'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 1,
            },
            "export": {
                "unit": "ug/m3",
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": aerosol.ConvertSulphate
        },
        {
            "regime": "IMG",
            "matrix": "ALL_AEROSOL",
            "comp_name": "sulphate_corrected",
            "ebas_unit": "ug S/m3",
            "import": {
                "units": ['ug/m3'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 1,
            },
            "export": {
                "unit": "ug/m3",
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": aerosol.ConvertSulphate
        },

        {
            "regime": "IMG",
            "matrix": "precip",
            "comp_name": "sulphate_total",
            "ebas_unit": "mg S/l",
            "import": {
                "units": ['mg/l', 'umol/l'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 1,
            },
            "export": {
                "unit": "mg/l",
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": precipitation.ConvertSulphate
        },
        {
            "regime": "IMG",
            "matrix": "precip",
            "comp_name": "sulphate_corrected",
            "ebas_unit": "mg S/l",
            "import": {
                "units": ['mg/l', 'umol/l'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 1,
            },
            "export": {
                "unit": "mg/l",
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": precipitation.ConvertSulphate
        },
        {
            "regime": "IMG",
            "matrix": "precip",
            "comp_name": "nitrate",
            "ebas_unit": "mg N/l",
            "import": {
                "units": ['mg/l', 'umol/l'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 1,
            },
            "export": {
                "unit": "mg/l",
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": precipitation.ConvertNitrate
        },
        {
            "regime": "IMG",
            "matrix": "precip",
            "comp_name": "ammonium",
            "ebas_unit": "mg N/l",
            "import": {
                "units": ['mg/l', 'umol/l'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 1,
            },
            "export": {
                "unit": "mg/l",
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": -1,
                "maxround": 0,
            },
            "cvt_class": precipitation.ConvertAmmonium
        },
        {
            "regime": "IMG",
            "matrix": "precip",
            "comp_name": "bicarbonate",
            "ebas_unit": "mg/l",
            "import": {
                "units": ['umol/l'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 1,
            },
            # no export conversion
            "cvt_class": precipitation.ConvertBicarbonate
        },
        {
            "regime": "IMG",
            "matrix": "precip",
            "comp_name": "calcium",
            "ebas_unit": "mg/l",
            "import": {
                "units": ['umol/l'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 1,
            },
            # no export conversion
            "cvt_class": precipitation.ConvertCalcium
        },
        {
            "regime": "IMG",
            "matrix": "precip",
            "comp_name": "chloride",
            "ebas_unit": "mg/l",
            "import": {
                "units": ['umol/l'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 1,
            },
            # no export conversion
            "cvt_class": precipitation.ConvertChloride
        },
        {
            "regime": "IMG",
            "matrix": "precip",
            "comp_name": "fluoride",
            "ebas_unit": "mg/l",
            "import": {
                "units": ['umol/l'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 1,
            },
            # no export conversion
            "cvt_class": precipitation.ConvertFluoride
        },
        {
            "regime": "IMG",
            "matrix": "precip",
            "comp_name": "magnesium",
            "ebas_unit": "mg/l",
            "import": {
                "units": ['umol/l'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 1,
            },
            # no export conversion
            "cvt_class": precipitation.ConvertMagnesium
        },
        {
            "regime": "IMG",
            "matrix": "precip",
            "comp_name": "phosphate",
            "ebas_unit": "mg/l",
            "import": {
                "units": ['umol/l'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 1,
            },
            # no export conversion
            "cvt_class": precipitation.ConvertPhosphate
        },
        {
            "regime": "IMG",
            "matrix": "precip",
            "comp_name": "potassium",
            "ebas_unit": "mg/l",
            "import": {
                "units": ['umol/l'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 1,
            },
            # no export conversion
            "cvt_class": precipitation.ConvertPotassium
        },
        {
            "regime": "IMG",
            "matrix": "precip",
            "comp_name": "sodium",
            "ebas_unit": "mg/l",
            "import": {
                "units": ['umol/l'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                "roundoffset": 1,
            },
            # no export conversion
            "cvt_class": precipitation.ConvertSodium
        },
        {
            "regime": "IMG",
            "matrix": "precip",
            "comp_name": "conductivity",
            "ebas_unit": "uS/cm",
            "import": {
                "units": ['mS/m'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                # no roundoffset and maxround, conversion is just factor 1000
            },
            # no export conversion
            "cvt_class": electric.ConvertConductivity
        },

        {
            "regime": "IMG",
            "matrix": "met",
            "comp_name": "temperature",
            "ebas_unit": "deg C",
            "import": {
                "units": ['K'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                # no roundoffset and maxround, conversion is just an addition
            },
            # no export conversion
            "cvt_class": ConvertTemperature
        },
        {
            "regime": "IMG",
            "matrix": "instrument",
            "comp_name": "temperature",
            "ebas_unit": "K",
            "import": {
                "units": ['deg C'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                # no roundoffset and maxround, conversion is just an addition
            },
            # no export conversion
            "cvt_class": ConvertTemperature
        },
        {
            "regime": "IMG",
            "matrix": "instrument",
            "comp_name": "pressure",
            "ebas_unit": "hPa",
            "import": {
                "units": ['Torr'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                # no roundoffset and maxround, conversion is just an addition
            },
            # no export conversion
            "cvt_class": ConvertPressure
        },

        {
            "regime": "IMG",
            "matrix": "instrument",
            "comp_name": "electric_current",
            "ebas_unit": "A",
            "import": {
                "units": ['mA'],
                "need_stdcond": False,
                "use_stdcond": (None, None),
                "set_stdcond": None,
                # no roundoffset and maxround, conversion is just an addition
            },
            # no export conversion
            "cvt_class": electric.ConvertElectricCurrent
        }
    ]

    CONVERSIONS_CHECKED = False

    def __init__(self):
        """
        Set up the object.
        Parameters:
            None
        Returns:
            None
        """
        self._check_conversions()

    @classmethod
    def converted_units(cls):
        """
        List all units which could be converted on import.
        This method is used for checking valid units in ebas.io
        """
        converted_units = []
        for cvt in cls.CONVERSIONS:
            if 'import' in cvt:
                converted_units += cvt['import']['units']
        return converted_units

    def _check_conversions(self):
        """
        Sanity checks for the CONVERSIONS configuration.
         - Check if parameter exists
         - Check if ebas_unit is correct
         - Check if conversion class name is correct
         - Check if masterdata.pm has exceptional masterdata for the converted
           unit on export: If not checked here, an exception will only be raised
           when the conversion in question is actually used on an export. Thus
           bugs might be hidden for a long time and then crash the export
           process at inconvenient times.
        Here we precheck if the masterdata are in place, raising an exception
        even if the conversion is not nedded now.
        Parametrs:
            None
        Returns:
            None
        Raises:
            KeyError if masterdata are not found
            RuntimeError if ebas_unit is not found
        """

        def generic_cvt_name(comp):
            """
            Generate a generic class name for conversion objects.
            Generic class names are 'Convert' + camel case on - and _
            Parameters:
                component name
            Returns:
                the generic name for a conversion name (convention)
            """
            parts = ('Convert ' + re.sub(r"(_|-)+", " ", comp)).split()
            return "".join([part[0].upper()+part[1:] for part in parts])

        if self.__class__.CONVERSIONS_CHECKED:
            return
        mpm = EbasMasterPM()

        # replace ALL_AEROSOL matrix with all defined matices:
        i = 0
        while i < len(self.__class__.CONVERSIONS):
            cvt = self.__class__.CONVERSIONS[i]
            if cvt['matrix'] == "ALL_AEROSOL":
                mats = mpm.list_matrix_group(cvt['regime'], cvt['matrix'],
                                             cvt['comp_name'])
                if not mats:
                    raise RuntimeError(
                        "No areosol parameters for conversion {}".format(cvt))
                for mat in mats:
                    new = cvt.copy()
                    new['matrix'] = mat
                    self.__class__.CONVERSIONS.append(new)
                del self.__class__.CONVERSIONS[i]
                # del element, do not incremet i
            else:
                i += 1

        for cvt in self.__class__.CONVERSIONS:
            # check existence of original parameter
            # (e.g. comp name renamed in ebas?)
            pm_ = mpm[(cvt['regime'], cvt['matrix'], cvt['comp_name'],
                       'arithmetic mean')]
            # raise KeyError if not extsts
            # (e.g. component renamed in the database)
            if pm_.PM_UNIT != cvt['ebas_unit']:
                raise RuntimeError("PM unit error ({})".format(cvt))

            # check class name for conversion class (avoid copy/paste errors,
            # make sure the class names are up to date when component names are
            # changed in ebas):
            if cvt['cvt_class'].__name__ != generic_cvt_name(cvt['comp_name']):
                if cvt['comp_name'] in \
                       ('nitrogen_monoxide', 'nitrogen_dioxide') and \
                   cvt['cvt_class'].__name__ == 'ConvertNOx':
                    pass  # exception OK
                elif cvt['comp_name'] in \
                       ('sulphate_total', 'sulphate_corrected') and \
                   cvt['cvt_class'].__name__ == 'ConvertSulphate':
                    pass  # exception OK
                else:
                    raise RuntimeError(
                        "Unconventional conversion class name '{}' "
                        "(should be '{}'?)".format(
                            cvt['cvt_class'].__name__,
                            generic_cvt_name(cvt['comp_name'])))

            # check exceptional PM for converted unit on export:
            # raises KeyError if not exists
            if 'export' in cvt:
                mpm.exceptional_unit_exp(
                    (cvt['regime'], cvt['matrix'], cvt['comp_name'],
                     cvt['export']['unit']))
        self.__class__.CONVERSIONS_CHECKED = True

    def import_conv_params(self, regime, matrix,  # pylint: disable=R0913
                           comp_name, unit, std_pres, std_temp):
        # R0913: Too many arguments
        # --> acceptable here
        """
        Get input converison object.
        Parameters:
            regime       }
            matrix       } source parameter definition
            comp_name    }
            unit         }
            std_pres     standard pressure from metadata (None if not set)
            std_temp     standard temperature from metadata (None if not set)
        Returns:
            tupel (conv_obj, set_stdcond)
                Conversion object (from atmos_phys.convert, e.g. ConvertNOx,
                    ConvertOzone etc).
                set_stscond: (volume standard pressure,
                              volume standard temperature) to be set after
                             conversion
                             None if no change is needed
        Raises:
            NoConversion if no conversion possible
        """
        msgs = []
        for cvt in self.__class__.CONVERSIONS:
            # loop through all defined coversions and choose if one is
            # applicable.
            # pylint: disable=R0916
            # R0916: Too many boolean expressions in if statement
            # --> not much to do about it?
            if 'import' in cvt and \
               cvt['regime'] == regime and \
               cvt['matrix'] == matrix and \
               cvt['comp_name'] == comp_name and \
               unit in cvt['import']['units']:
                if cvt['import']['need_stdcond'] and \
                       (std_pres is None or std_temp is None):
                    msgs.append(
                        ("Regime/Matrix/Component combination '{}'/'{}'/'{}': "
                         "unit '{}' cannot be converted to '{}' "
                         "(standard conditions missing)").format(
                             regime, matrix, comp_name, unit, cvt['ebas_unit']))
                else:
                    args = [unit, cvt['ebas_unit']]  # from_unit, to_unit
                    kwargs = {}
                    if "roundoffset" in cvt['import']:
                        kwargs["roundoffset"] = cvt['import']["roundoffset"]
                    if "maxround" in cvt['import']:
                        kwargs["maxround"] = cvt['import']["maxround"]
                    if cvt['import']['need_stdcond']:
                        kwargs["pressure"] = std_pres*100
                        kwargs["temperature"] = std_temp
                    elif cvt['import']['use_stdcond']:
                        if cvt['import']['use_stdcond'][0] is not None:
                            kwargs["pressure"] = \
                                cvt['import']['use_stdcond'][0] * 100
                        if cvt['import']['use_stdcond'][1] is not None:
                            kwargs["temperature"] = cvt['import']['use_stdcond'][1]
                    return (cvt['cvt_class'](*args, **kwargs),
                            cvt['import']['set_stdcond'])
        raise NoConversion(msgs)

    def export_conv_params(self, regime, matrix,  # pylint: disable=R0913
                           comp_name, unit, std_pres, std_temp):
        # R0913: Too many arguments
        # --> acceptable here
        """
        Get output converison object.
        Parameters:
            regime       }
            matrix       } source parameter definition
            comp_name    }
            unit         }
            std_pres     standard pressure from metadata (None if not set)
            std_temp     standard temperature from metadata (None if not set)
        Returns:
            tupel (conv_obj, set_stdcond)
                Conversion object (from atmos_phys.convert, e.g. ConvertNOx,
                    ConvertOzone etc).
                set_stscond: (volume standard pressure,
                              volume standard temperature) to be set after
                             conversion
                             None if no change is needed
        Raises:
            NoConversion if no conversion possible
        """
        msgs = []
        for cvt in self.__class__.CONVERSIONS:
            # loop through all defined coversions and choose if one is
            # applicable.
            # pylint: disable=R0916
            # R0916: Too many boolean expressions in if statement
            # --> not much to do about it?
            if 'export' in cvt and \
               cvt['regime'] == regime and \
               cvt['matrix'] == matrix and \
               cvt['comp_name'] == comp_name and \
               cvt['ebas_unit'] == unit:
                if cvt['export']['need_stdcond'] and \
                        (std_pres is None or std_temp is None):
                    msgs.append(
                        ("Regime/Matrix/Component combination '{}'/'{}'/'{}': "
                         "unit '{}' cannot be converted to '{}' "
                         "(standard conditions missing)").format(
                             regime, matrix, comp_name, unit,
                             cvt['export']['unit']))
                else:
                    args = [unit, cvt['export']['unit']]  # from_unit, to_unit
                    kwargs = {}
                    if "roundoffset" in cvt['export']:
                        kwargs["roundoffset"] = cvt['export']["roundoffset"]
                    if "maxround" in cvt['export']:
                        kwargs["maxround"] = cvt['export']["maxround"]
                    if cvt['export']['need_stdcond']:
                        kwargs["pressure"] = std_pres*100
                        kwargs["temperature"] = std_temp
                    elif cvt['export']['use_stdcond']:
                        if cvt['export']['use_stdcond'][0] is not None:
                            kwargs["pressure"] = \
                                cvt['export']['use_stdcond'][0] * 100
                        if cvt['export']['use_stdcond'][1] is not None:
                            kwargs["temperature"] = cvt['export']['use_stdcond'][1]
                    return (cvt['cvt_class'](*args, **kwargs),
                            cvt['export']['set_stdcond'])
        raise NoConversion(msgs)



"""
Xml/__init__.py
$Id: __init__.py 1198 2016-03-10 09:05:39Z pe $

XML module

History:
V.1.0.0  2013-04-15  pe  initial version

"""

from .base import XmlError
from .write import XmlWrite
from nilutility.datatypes import HexInt

class Xml(XmlWrite):
    """
    Partial class EbasXml: Put together the parts to the final class (only write, no read class).
    """
    pass


"""
ebas/input_output/ebas_xml/write.py
$Id: write.py 2519 2020-10-30 16:13:09Z pe $

output functionality for EBAS XML module

History:
V.1.0.0  2014-08-05  toh  initial version

"""

import sys
import os
from six import PY2
from .base import EbasXMLBase
from ..base import FLAGS_ONE_OR_ALL, FLAGS_NONE
from ...ebasmetadata import EbasMetadata
from ebas.domain.basic_domain_logic.time_period import estimate_period_code
from nilutility.datatypes import DataObject
from fileformats.Xml.ebas import Xml
from ...fileset.xmlwrap import xmlwrap_fileheader, xmlwrap_filetrailer
from ..base import EBAS_IOFORMAT_XML

class EbasXMLPartialWrite(EbasXMLBase):  # pylint: disable=R0901, W0223,
    # R0901: Too many ancestors
    # This is a quite big class, therefore we ignore this warning
    # W0223: ... is abstract in base class but is not overridden
    # this class is abstract as well (does not have any NotImplementedError)
    """
    XML I/O object.

    This is one part of the partial class EbasXML (output functionality).
    """

    def write(self, createfiles=False, destdir=None,  # pylint: disable=R0913
              xmlwrap=False, fileobj=None, flags=FLAGS_ONE_OR_ALL,
              datadef='EBAS_1.1', metadata_options=0, suppress=0):
        # R0913 Too many arguments
        # W0613 (unused metadata_options): need to be fixed? implement?
        """
        Writes the I/O object to an XML file (or stdout)
        Parameters:
            createfiles   write to file? else write to stdout
                          (special case for fileobj)
            destdir       destination directory path for files
            xmlwrap       wrap output in xml container
            fileobj       write to filehandle
            flags         flag column options: FLAGS_ONE_OR_ALL (default),
                          FLAGS_COMPRESS, FLAGS_ALL, FLAGS_NONE
            datadef       ebas data definition (EBAS_1 or EBAS_1.1)
            metadata_options
                          options for output, bitfield:
                          currently only EBAS_IOMETADATA_OPTION_SETKEY
                          (ignored in XML, setkey is written anyway)
            suppress      suppress selected consolidations (bitfield):
                              SUPPRESS_SORT_VARIABLES
                              SUPPRESS_METADATA_OCCURRENCE
        Returns:
            None
        """
        fil = self.prepare_write(
            True, createfiles=createfiles, destdir=destdir, xmlwrap=xmlwrap,
            fileobj=fileobj, flags=flags, datadef=datadef,
            metadata_options=metadata_options, suppress=suppress)
        if  createfiles:
            fil.write('<?xml version="1.0" encoding="utf-8"?>\n')
            fil.write('<ebas:Result xmlns:ebas="http://ebas.nilu.no/EBAS">\n')
        self.xml = Xml(self.result_id)
        self.set_attribs(flags=flags)
        self.set_times(flags=flags)
        self.set_data(flags=flags)
        self.set_flags(flags=flags)
        self.xml.write(fil)
        if  createfiles:
            fil.write('</ebas:Result>\n')
        if xmlwrap:
            xmlwrap_filetrailer()
        self.write_indexdb()

    def set_datadef(self, datadef):
        """
        Set the EBAS data definition (metadata standard).
        Parameters:
            datadef    data definition (EBAS_1 or EBAS_1.1)
        Returns
            None
        """
        # must be implemented here, because ebasmetadata can't be imported in
        # ..base
        self.metadata.datadef = datadef
        self.internal.ebasmetadata = EbasMetadata(
            self.metadata.datadef, data_format=EBAS_IOFORMAT_XML,
            data_level=self.metadata.datalevel)

    def set_attribs(self, flags=FLAGS_ONE_OR_ALL):
        """
        Sets directly the non-data Xml attributes.
        Parameters:
            None
        Returns
            None
        """
        self.xml.STATION_CODE = self.get_metadata_list_per_var('station_code')
        self.xml.STATION_NAME = self.get_metadata_list_per_var('station_name')
        self.xml.INST_TYPE = self.get_metadata_list_per_var('instr_type')
        self.xml.COMP_NAME = self.get_metadata_list_per_var('comp_name')
        self.xml.MATRIX_NAME = self.get_metadata_list_per_var('matrix')
        self.xml.UNIT = self.get_metadata_list_per_var('unit')
        self.xml.STATISTIC_CODE = self.get_metadata_list_per_var('statistics')
        self.xml.RESCODE = self.get_metadata_list_per_var('resolution')
        self.xml.DATA_LEVEL = self.get_metadata_list_per_var('datalevel')
        self.xml.HEIGHT_AGL = self.get_metadata_list_per_var('mea_height')
        self.xml.SETKEY = self.get_metadata_list_per_var('setkey')

        self.xml.UNCERTAINTY = self.get_metadata_list_per_var('uncertainty') # tupel

        self.xml.POSITION = list(zip(self.get_metadata_list_per_var('station_latitude'),
                                self.get_metadata_list_per_var('station_longitude'),
                                self.get_metadata_list_per_var('station_altitude')))

        # dictionary
        self.xml.CHARACTERISTICS = self.get_metadata_list_per_var('characteristics')

    def set_times(self, flags=FLAGS_ONE_OR_ALL):
        """
        Set time attributes.
        Parameter:
            None
        Returns:
            None
        """
        self.xml.SAMPLE_TIMES = self.sample_times

    def set_data(self, flags=FLAGS_ONE_OR_ALL):
        """
        Sets data attributes.
        Parameter:
            None
        Returns:
            None
        """
        self.xml.DATA = []
        for j in range(len(self.variables)):
            self.xml.DATA.append(self.values_without_flags(j))
            # clean values (set invalid to missing, set 781 to 1/2 value)
            # is done only for export - so the XML object remains
            # fully valid!

    def set_flags(self, flags=FLAGS_ONE_OR_ALL):
        """
        Sets data attributes.
        Parameter:
            None
        Returns:
            None
        """
        self.xml.FLAGS = []
        for j in range(len(self.variables)):
            if self.variables[j].flagcol:
                if flags == FLAGS_NONE:
                    # clean values (set invalid to missing, set 781 to 1/2
                    # value) is done only for export - so the NasaAmes object
                    # remains fully valid!
                    self.xml.FLAGS.append('')
                else:
                    flagstrings = [','.join(["{0:03d}".format(flag)
                                             for flag in flags])
                                   for flags in self.variables[j].flags]
                    self.xml.FLAGS.append(flagstrings)

    def join(self, other):
        """
        Join self with other if possible.
        Default implementation: do not join
        Parameters:
            other    other EbasIO object
        Returns
            New, joined EbasIO object
            None (no join possible)
        """
        # check if objects can be joined:
        # same variables, same setkeys, data interval can be joined,
        # and ALL metadata that are reported in the XML format must be the same:
        #    mea_height
        # Most depend on setkey (DS or HDS) and do not need to be checked:
        #  station_code
        #  station_name
        #  instr_type
        #  comp_name
        #  matrix
        #  unit
        #  statistics
        #  resolution
        #  datalevel
        #  characteristics
        if len(self.variables) != len(other.variables) or \
           self.get_metadata_list_per_var('setkey') != \
               other.get_metadata_list_per_var('setkey') or \
           self.sample_times[-1][1] > other.sample_times[0][0] or \
           self.get_metadata_list_per_var('mea_height') != \
               other.get_metadata_list_per_var('mea_height') or \
           self.get_metadata_list_per_var('uncertainty') != \
               other.get_metadata_list_per_var('uncertainty'):
            return None
        from .xml import EbasXML
        new = EbasXML(result_id=self.result_id, indexdb=self.indexdb)
        metadata = ['station_code', 'station_name', 'station_latitude',
                    'station_longitude', 'station_altitude']
        for meta in metadata:
            if self.metadata[meta] is not None:
                new.metadata[meta] = self.metadata[meta]
        new.sample_times = self.sample_times + other.sample_times
        new.variables = []
        for vnum in range(len(self.variables)):
            var1 = self.variables[vnum]
            var2 = other.variables[vnum]
            new.variables.append(
                DataObject(
                    is_hex=var1.is_hex,
                    values_=var1.values_ + var2.values_,
                    flags=var1.flags + var2.flags,
                    flagcol=True,
                    metadata=DataObject(
                        setkey=var1.metadata.setkey,
                        comp_name=var1.metadata.comp_name,
                        unit=var1.metadata.unit,
                        )))
            char = self.get_characteristics_for_var(vnum)
            if char:
                new.variables[vnum].metadata.characteristics = char
            metadata = [
                'instr_type', 'matrix', 'statistics', 'datalevel',
                'characteristics', 'resolution', 'mea_height', 'uncertainty',
                # plus some that are used for filename generation
                # (must be the same in both because they only depend on DS)
                'lab_code', 'instr_name', 'method']
            for meta in metadata:
                val = self.get_meta_for_var(vnum, meta)
                if val is not None:
                    new.variables[vnum].metadata[meta] = val
        return new



"""
EBAS domain layer (entity objects and domain logic)

The ebas domain module implements the core domain logic in EBAS.

 * ::: ebas.domain.entities
       options:
           members: no
 * ::: ebas.domain.access_statistics
       options:
           members: no
 * ::: ebas.domain.masterdata
       options:
           members: no
"""

from .base import EbasDomainError, EbasDomainDbContext



"""
$Id: dataset_types.py 2800 2022-06-09 11:20:46Z pe $

Basic functionality for classifying dataset types
(auxiliary, precipitation amount)

Those methods are used both in ebas.io and ebas.domain.entities.ds.
In order to make them available for both modules they are separated in this
general domain module.
"""

def is_precip_concentration(matrix, comp_name):
    """
    Check if the dataset is a precipitation concentration component (one
    that needs precipitation amount)
    Parameters:
        comp_name   component name
        matrix      matrix name
    Returns:
        True/False
    """
    if matrix in ('precip', 'precip_tot') and \
            not comp_name.startswith('precipitation_amount'):
        return True
    return False

def is_precip_amount(matrix, comp_name):
    """
    Check if the dataset is a precipitation amount component
    Parameters:
        comp_name   component name
        matrix      matrix name
    Returns:
        True/False
    """
    if matrix in ('precip', 'precip_tot') and \
            comp_name.startswith('precipitation_amount'):
        return True
    return False

def is_auxiliary(matrix, comp_name):
    """
    Check if the dataset is to be considered "auxiliary data".
    Parameters:
        comp_name   component name
        matrix      matrix name
    Returns:
        True/False
    """
    if matrix in ('instrument', 'position'):
        return True
    if matrix != 'met' and \
       comp_name in  ('pressure', 'temperature', 'relative_humidity'):
        return True
    if comp_name == 'sample_count':
        # Component sample count is auxiliary.
        # But parameters with statistics=sample count are not!
        return True
    if comp_name in ('start_time', 'end_time', 'number of size bins'):
        # These are variables skipped on import, but when creating files, they
        # are of course auxiliary.
        return True
    return is_precip_amount(matrix, comp_name)



#!/usr/bin/env python
"""
update.py
$Id: update.py 2743 2021-11-23 15:48:15Z pe $

Update offline masterdata from database.

History:
V.0.1.0  2014-02-28  pe  initial version

"""
__version__ = '0.1.0-devel $Rev: 2743 $'


import logging

# if __name__ == '__main__':
#     if __package__ is None:
import sys
from os import path
#         sys.path.append(path.dirname(path.dirname(path.abspath(__file__))))
sys.path.append(path.join(path.dirname(path.abspath(__file__)),
                          '..', '..', '..', '..'))
#         import offline_masterdata
#     else:
#         from . import offline_masterdata
# # need some path tweaking, add ../.. (= ebas lib) to python path

import offline_masterdata
from ebas.commandline import EbasCommandline

def update_masterdata(cmdline):
    """
    Main program for update
    Created for lexical scoping.

    Parameters:
        none
    Returns:
        none
    """
    logger = logging.getLogger('update')

    dbh = cmdline.dbh
    # don't register user with db layer, not necessary for update masterdata

    logger.info(' AC')
    offline_masterdata.ACOfflineMasterData.write_pickle_file(dbh)
    logger.info(' AX')
    offline_masterdata.AXOfflineMasterData.write_pickle_file(dbh)
    logger.info(' AM')
    offline_masterdata.AMOfflineMasterData.write_pickle_file(dbh)
    logger.info(' AY')
    offline_masterdata.AYOfflineMasterData.write_pickle_file(dbh)
    logger.info(' BC')
    offline_masterdata.BCOfflineMasterData.write_pickle_file(dbh)
    logger.info(' CA')
    offline_masterdata.CAOfflineMasterData.write_pickle_file(dbh)
    logger.info(' CC')
    offline_masterdata.CCOfflineMasterData.write_pickle_file(dbh)
    logger.info(' CO')
    offline_masterdata.COOfflineMasterData.write_pickle_file(dbh)
    logger.info(' CT and CV')
    offline_masterdata.CTOfflineMasterData.write_pickle_file(dbh)
    logger.info(' CY')
    offline_masterdata.CYOfflineMasterData.write_pickle_file(dbh)
    logger.info(' DL')
    offline_masterdata.DLOfflineMasterData.write_pickle_file(dbh)
    logger.info(' FL')
    offline_masterdata.FLOfflineMasterData.write_pickle_file(dbh)
    logger.info(' FM')
    offline_masterdata.FMOfflineMasterData.write_pickle_file(dbh)
    logger.info(' FT')
    offline_masterdata.FTOfflineMasterData.write_pickle_file(dbh)
    logger.info(' FP')
    offline_masterdata.FPOfflineMasterData.write_pickle_file(dbh)
    logger.info(' HT')
    offline_masterdata.HTOfflineMasterData.write_pickle_file(dbh)
    logger.info(' IM')
    offline_masterdata.IMOfflineMasterData.write_pickle_file(dbh)
    logger.info(' IP')
    offline_masterdata.IPOfflineMasterData.write_pickle_file(dbh)
    logger.info(' IT')
    offline_masterdata.ITOfflineMasterData.write_pickle_file(dbh)
    logger.info(' MA')
    offline_masterdata.MAOfflineMasterData.write_pickle_file(dbh)
    logger.info(' MD')
    offline_masterdata.MDOfflineMasterData.write_pickle_file(dbh)
    logger.info(' OC')
    offline_masterdata.OCOfflineMasterData.write_pickle_file(dbh)
    logger.info(' OR')
    offline_masterdata.OROfflineMasterData.write_pickle_file(dbh)
    logger.info(' PM')
    offline_masterdata.PMOfflineMasterData.write_pickle_file(dbh)
    logger.info(' PG')
    offline_masterdata.PGOfflineMasterData.write_pickle_file(dbh)
    logger.info(' PL')
    offline_masterdata.PLOfflineMasterData.write_pickle_file(dbh)
    logger.info(' PR')
    offline_masterdata.PROfflineMasterData.write_pickle_file(dbh)
    logger.info(' PS')
    offline_masterdata.PSOfflineMasterData.write_pickle_file(dbh)
    logger.info(' QM')
    offline_masterdata.QMOfflineMasterData.write_pickle_file(dbh)
    logger.info(' QO')
    offline_masterdata.QOOfflineMasterData.write_pickle_file(dbh)
    logger.info(' QV')
    offline_masterdata.QVOfflineMasterData.write_pickle_file(dbh)
    logger.info(' RE')
    offline_masterdata.REOfflineMasterData.write_pickle_file(dbh)
    logger.info(' SE')
    offline_masterdata.SEOfflineMasterData.write_pickle_file(dbh)
    logger.info(' SM')
    offline_masterdata.SMOfflineMasterData.write_pickle_file(dbh)
    logger.info(' SC')
    offline_masterdata.SCOfflineMasterData.write_pickle_file(dbh)
    logger.info(' SP')
    offline_masterdata.SPOfflineMasterData.write_pickle_file(dbh)
    logger.info(' ST')
    offline_masterdata.STOfflineMasterData.write_pickle_file(dbh)
    logger.info(' SX')
    offline_masterdata.SXOfflineMasterData.write_pickle_file(dbh)
    logger.info(' SZ')
    offline_masterdata.SZOfflineMasterData.write_pickle_file(dbh)
    logger.info(' SW')
    offline_masterdata.SWOfflineMasterData.write_pickle_file(dbh)
    logger.info(' VP')
    offline_masterdata.VPOfflineMasterData.write_pickle_file(dbh)
    logger.info(' VT')
    offline_masterdata.VTOfflineMasterData.write_pickle_file(dbh)
    logger.info(' WC')
    offline_masterdata.WCOfflineMasterData.write_pickle_file(dbh)
    logger.info(' ZN')
    offline_masterdata.ZNOfflineMasterData.write_pickle_file(dbh)
    logger.info(' ZT')
    offline_masterdata.ZTOfflineMasterData.write_pickle_file(dbh)
    logger.info('all masterdata caches updated')


EbasCommandline(update_masterdata, custom_args=['LOGGING', 'DB', 'CONFIG'],
                help_description='%(prog)s update masterdata caches.',
                version=__version__, read_masterdata=False).run()



"""
ebas/domain/masterdata/pm.py
$Id: pg_pl.py 2502 2020-10-14 16:57:35Z pe $

EBAS Masterdata Class for parameter masterdata

This module implements the class EbasMasterPG and PL.

History:
V.1.0.0  2014-10-06  pe  initial version

"""

from .offline_masterdata import PGOfflineMasterData, PLOfflineMasterData
from .base import EbasMasterBase


def uniq_list(lst):
    """
    List with only unique elements, keep order.
    """
    x = set()
    return [0 if x.add(y) else y for y in lst if y not in x]


class EbasMasterPG(EbasMasterBase, PGOfflineMasterData):
    """
    Domain Class for Parameter group masterdata.
    Objects of this class do not represent entities, this class provides class
    methods for handling parameter and checking them against master data.
    Parameter master data are retrieved from database or from offline storage
    when no database access is possible.
    """

    # read offline masterdata for PG (parameter groups)
    # Those are fallback values, will be read from database as soon as possible.
    PGOfflineMasterData.read_pickle_file()

    def __init__(self, dbh=None):
        """
        Read pg masterdata if dbh is provided.
        """
        PGOfflineMasterData.__init__(self)
        if not self.__class__.INIT and dbh:
            self.from_db(dbh)


class EbasMasterPL(EbasMasterBase, PLOfflineMasterData):
    """
    Domain Class for Parameter group list masterdata.
    Objects of this class do not represent entities, this class provides class
    methods for handling parameter and checking them against master data.
    Parameter master data are retrieved from database or from offline storage
    when no database access is possible.
    """

    # read offline masterdata for PL (parameter group list)
    # Those are fallback values, will be read from database as soon as possible.
    PLOfflineMasterData.read_pickle_file()

    def __init__(self, dbh=None):
        """
        Read PL masterdata if dbh is provided.
        """
        PLOfflineMasterData.__init__(self)
        if not self.__class__.INIT and dbh:
            self.from_db(dbh)
        self._pg = EbasMasterPG(dbh=dbh)

    def list_groups(self, regime, matrix, comp_name):
        """
        List all groups for a given parameter
        Order the groups by descriptivity (i.e. comp_no ascending). Groups with
        many components are less descriptive than groups with less components.

        Does not use all_air and all_precip!
        Those are too wide and not descriptive.
        """
        raw = [(elem['PG_NAME'], self._pg[elem['PG_NAME']]['CO_COUNT'])
                for grp in self.META for elem in self.META[grp]
                if (elem['RE_REGIME_CODE'] is None or
                    regime == elem['RE_REGIME_CODE']) and \
                   (elem['MA_MATRIX_NAME'] is None or
                    matrix == elem['MA_MATRIX_NAME']) and \
                   (elem['CO_COMP_NAME'] is None or
                    comp_name == elem['CO_COMP_NAME'])]
        return [elem[0]
                for elem in sorted(raw, key=lambda x: x[1])
                if elem[0] not in ('all_air', 'all_precip')]

    def list_group_desc(self, regime, matrix, comp_name):
        """
        List all groups for a given parameter - prefer group desc if set, else
        name
        Order the groups by descriptivity (i.e. comp_no ascending). Groups with
        many components are less descriptive than groups with less components.
        """
        def pref_desc(name):
            """
            Returns the prefered description if exists, else the name
            """
            return (self._pg[name]['PG_DESC'] if self._pg[name]['PG_DESC']
                    else name)

        return uniq_list(
            [pref_desc(grp)
             for grp in self.list_groups(regime, matrix, comp_name)])



"""
CDM reader
"""

import json
import requests
import re
import datetime
import functools
from xarray import open_dataset, backends
import logging
import os
from nilutility.datetime_helper import DatetimeInterval

SOURCE_NETCDF = 1
SOURCE_OPENDAP = 2
OPENDAP_BACKEND_NETCDF = 1
OPENDAP_BACKEND_PYDAP = 2

functools.lru_cache(maxsize=10)
def lrucache_get_time_dependent_metadata(self, var):
    """
    LRU chache for EbasCDMReader.get_time_dependent_metadata().
    """
    return EbasCDMReader._get_time_dependent_metadata(self, var)
    
class EbasCDMReader(object):
    """
    CDM reader class
    """

    def __init__(self, resource, opendap_backend=None):
        """
        Initialize CDM reader
        Parameters:
            resource: file path or OPeNDAP url
        """
        self.logger = logging.getLogger(__name__)
        self.resource = resource
        self.dataset = None
        if os.path.isfile(resource):
            self.logger.debug("Open local NetCDF file: %s", resource)
            self.source_type = SOURCE_NETCDF
            self.dataset = open_dataset(resource)
        elif resource.startswith('http'):
            self.source_type = SOURCE_OPENDAP
            self.backend_type = None
            self._open_opendap(resource, backend=opendap_backend)
        else:
            self.logger.error(f"Resource {resource} is not valid.")
            raise ValueError(f"Resource {resource} is not valid.")        

    def _open_opendap(self, url, backend=None):
        """
        Open the OPeNDAP url.
        
        There is a problem opening an OPeNDAP url with xarray using the
        NetCDF backend for some files/variables (many variables).

        An alternative is to use the pydap backend via xarray, but this is
        slower than the NetCDF backend.

        Caller can either choose a specific backend or use auto selection:
        In this case NetCDF is tried first, then pydap.
        The caller may use EbasCDMReader.backend_type to check which backend was
        used (e.g. for implementing some learning).

        Parameters:
            url: OPeNDAP url
            backend: backend to use (OPENDAP_BACKEND_NETCDF,
                     OPENDAP_BACKEND_PYDAP or None for auto selection)
            Returns:
                None
        """

        def _open_pydap(url):
            store = backends.PydapDataStore.open(
                url, user_charset='UTF-8')
            return open_dataset(store)

        self.logger.debug("Open OPeNDAP url: %s", url)
        if backend == OPENDAP_BACKEND_PYDAP:
            # force pydap backend
            self.dataset =_open_pydap(url)
            self.backend_type = OPENDAP_BACKEND_PYDAP
            return
        elif backend == OPENDAP_BACKEND_NETCDF:
            # force NetCDF backend
            self.dataset = open_dataset(url)
            self.backend_type = OPENDAP_BACKEND_NETCDF
            return
        # auto select backend, try NetCDF first
        try:
            self.dataset = open_dataset(url)
            self.backend_type = OPENDAP_BACKEND_NETCDF
            return
        except RuntimeError:
            # use pydap fallback
            self.dataset = _open_pydap(url)
            self.backend_type = OPENDAP_BACKEND_PYDAP
            return

    def get_time_dependent_metadata(self, var):
        """
        Get time dependent metadata
        Parameters:
            var: variable name of the base variable
        Returns:
            dict with time dependent metadata
        This uses a LRU cache, the real get is implemented in
        _get_time_dependent_metadata().
        """
        return lrucache_get_time_dependent_metadata(self, var)

    def _get_time_dependent_metadata(self, var):
        """
        
        """
        var = var + '_ebasmetadata'
        if self.source_type == SOURCE_NETCDF:
            ret = []
            for txt in self.dataset[var][:]:
                ret.aapend(json.loads(txt))
        else:  # OPeNDAP:
            # There is a problem with reading the string arrays via xarray and
            # OPeNDAP (independent of the backend used). Only the first 64
            # characters are returned. This makes the json invalid.
            # As a workaround, the metadata variables are read via an OPenDAP
            # get ascii request manually, then converted to a clean json string and
            # parsed.
            # A note on the dods request: This had anoter issue with charset
            # encoding:
            #   The test only includes on NRT dataset, so the test case can not be
            #   referenced by DOI currently, here's a long description:
            #   The test case included PS 17642 (address is "Chemin de l’Aeorologie"
            #   including \xe2\x80\x99 (unicode U+2019 RIGHT SINGLE QUOTATION MARK)
            #   (this character has no iso-1 representation).
            #   The test case also included organisation CH08L (address
            #   "Chemin de l'Aérologie" with a normal ascii single quotation mark
            #   (0x27), and additionally an é character (\xc3\xa9, unicode U+00E9)
            #   LATIN SMALL LETTER E WITH ACUTE. This character is representable in
            #   iso-1 (E9).
            # The DODS, e.g https://thredds.nilu.no/thredds/dodsC/auto-pollen_nrt/CH0001G.20240701100000.20251020235427.pollen_monitor..aerosol.68w.1h.CH08L_JUPITER_CH0001G_1_NRT.FI01L_FMI_Poleno.lev1.5.nc.dods?pollen_alnus_amean_ebasmetadata)
            # The response conent contains:
            #   for the institute b"Chemin de l\'A\xe9rologie" -- the é is iso-1 encoded
            #   for the originator b"Chemin de l?Aeorologie" -- the U+2019 is replaced by ?
            # Conclusion: the DODS request returns iso-1 encoded content, unicode
            # characters not representable in iso-1 are lost (replaced by ?).
            # --> DODS is not useable for metadata.
            #
            # Last resort: the ascii request is used, which returns utf-8 content.
            
            url = self.resource + '.ascii?' + var
            res = requests.get(url)
            if res.status_code != 200:
                self.logger.error(
                    "Failed to fetch metadata via OPeNDAP ascii request: %s",
                    url)
                raise FileNotFoundError(
                    "Failed to fetch metadata via OPeNDAP ascii request")
            txt=res.content.decode('utf8')
            txt = re.sub('^Dataset.*---------------------------------------------\n[^\n]*\n"', '[', txt, flags=re.DOTALL)
            txt = re.sub('"\n}", "{\n', '"\n},\n{\n', txt, count=0, flags=re.DOTALL+re.MULTILINE)
            txt = re.sub('"\n*$', '\n]', txt)
            ret = json.loads(txt)
        # add metadata time range to the metadata entries
        for i, times in enumerate(self.dataset['metadata_time_bnds'][:]):
            ret[i]['time_range'] = DatetimeInterval(
                times[0].values.astype('datetime64[us]').astype(
                    datetime.datetime).replace(
                        tzinfo=datetime.timezone.utc),
                times[1].values.astype('datetime64[us]').astype(
                    datetime.datetime).replace(
                        tzinfo=datetime.timezone.utc))
        return ret



"""
/Xml/base.py
$Id: base.py 2561 2020-12-07 23:09:30Z pe $

Base class for XML.
Has (still) all attributes as base class for Nasa Ames 1001.

History:
V.1.0.0  2014-08-05  toh  initial version

"""
import logging
from nilutility.ndecimal import Decimal
from nilutility.float_helper import float_almostequal
from nilutility.datatypes import DataObject

class XmlError(Exception):
    """
    Error class for XML.
    """
    pass
    
class XmlBase(object):
    # pylint: disable-msg=R0902
    # R0902: Too many instance attributes
    """
    Partial class Xml: Base class used by Parts.
    """

    # all data attibutes (set on read, getattr implemented)
    # toh: all attributes are lists as every variable will be
    #      written into a separate result set; thus attributes
    #      might vary with variables
    #      (although it's not necessary/implemented for every attribute yet)
    DATA_ATTRIBS = [
        'SAMPLE_TIMES',     # list of time tuples; local: sample_times
        'STATION_CODE',     # list of str; XML: StationCode,    EBAS: ST_STATION_CODE,   local: [variables[vnum].]metadata.station_code
        'STATION_NAME',     # list of str; XML: StationName,    EBAS: ST_NAME,           local: [variables[vnum].]metadata.station_name
        'INST_TYPE',        # list of str; XML: InstrumentType, EBAS: FT_TYPE,           local: [variables[vnum].]metadata.instr_type
        'COMP_NAME',        # list of str; XML: ComponentName,  EBAS: CO_COMP_NAME,      local: [variables[vnum].]metadata.comp_name
        'CHARACTERISTICS',  # list of dic; XML: Characteristics EBAS: CT_TYPE ,          local: [variables[vnum].]metadata.characteristics
        'MATRIX_NAME',      # list of str; XML: Matrix,         EBAS: MA_MATRIX_NAME,    local: [variables[vnum].]metadata.matrix
        'UNIT',             # list of str; XML: Unit,           EBAS: PM_UNIT,           local: [variables[vnum].]metadata.unit
        'STATISTIC_CODE',   # list of str; XML: Statistics,     EBAS: SC_STATISTIC_CODE, local: [variables[vnum].]metadata.statistics
        'RESCODE',          # list of str; XML: ResolutionCode, EBAS: DS_RESCODE,        local: [variables[vnum].]metadata.resolution
        'DATA_LEVEL',       # list of str; XML: DataLevel,      EBAS: DL_DATA_LEVEL,     local: [variables[vnum].]metadata.datalevel; optional
        'UNCERTAINTY',      # list of str; XML: MeasurementUncertainty, EBAS: HVM_UNCERTAINTY, local: [variables[vnum].]metadata.uncertainty
        'POSITION',         # list of str; XML: Position,       EBAS: ST_LATITUDE, ST_LONGITUDE, ST_ALTITUDE_ASL; local: station_latitude, station_longitude, station_altitude
        'HEIGHT_AGL',       # list of str; XML: HeightAGL,      EBAS: HVM_HEIGHT_AGL,    local: [variables[vnum].]metadata.mea_height; optional
        'DATA',             # list of lists of float/decimal; XML: Values
        'SETKEY',           # list of str; XML: setkey,         EBAS: DS_SETKEY          local: [variables[vnum].]metadata.setkey
        'FLAGS',            # list of lists of str;             toh: flags are not required yet
        'NV',               # int;           number of primary dependent variables
        # used by write only:
        ]

    # attributes that can be set (before write), others will be calculated
    SETABLE_DATA_ATTRIBS = [
                            'SAMPLE_TIMES',
                            'STATION_CODE',
                            'STATION_NAME',
                            'INST_TYPE',
                            'COMP_NAME',
                            'CHARACTERISTICS',
                            'MATRIX_NAME',
                            'UNIT',
                            'STATISTIC_CODE',
                            'RESCODE',
                            'DATA_LEVEL',
                            'UNCERTAINTY',
                            'POSITION',
                            'HEIGHT_AGL',
                            'DATA',
                            'SETKEY',
                            'FLAGS'
                            ]
    # read/write states
    # toh: ignore all items that are not required by XML
    # SAMPLE_TIMES will be written by default
    SECTIONS = [
                # 'SAMPLE_TIMES', # will be written by default
                'STATION_CODE',
                'STATION_NAME',
                'INST_TYPE',
                'COMP_NAME',
                'CHARACTERISTICS',
                'MATRIX_NAME',
                'UNIT',
                'STATISTIC_CODE',
                'RESCODE',
                'DATA_LEVEL',
                'UNCERTAINTY',
                'POSITION',
                'HEIGHT_AGL',
                'DATA',
                'SETKEY'
                ]
    
    # for bitfield strictness:
    # strictly require data in format given by VMISS
    STRICT_NUMFORMAT = 1 
    STRICT_DX = 2
    # strictly require irregular data when DX=0 (DX!=0 will always raise an
    # error if not obeyed in data)
    
    def __setattr__(self, name, value):
        if name in self.__class__.SETABLE_DATA_ATTRIBS:
            self.data[name] = value
        else:
            object.__setattr__(self, name, value)

    def __getattr__(self, name):
        try:
            return self.data[name]
        except KeyError as expt:
            raise AttributeError(expt)

    def __init__(self, resultset_id, strictness=0):
        """
        Paramaters:
            resultset_id    increasing number (id for the resultset to be
                            generated) - needed only in _set_pos()
            strictness      bit field for strictness on read
                            (currently only STRICT_NUMFORMAT)
        Returns:
            None
        Raises:
            IOError on file open problems
            NasaAmesError
        """
        self.file = None
        self.state = -1
        self.errors = 0
        self.warnings = 0
        self.resultset_id = resultset_id
        self.strictness = strictness
        self.lnum = 0
        # some special attributes:
        # xformat: calcuated and used internally by write:
        self.xformat = None
        # xmindig: minimum digit of x used in the physical file
        # (order of magnitude)
        # is set by the reader when reading the data lines
        self.xmindig = None
        # min/max xmindig: can be set by the caller to restrict the output
        # precision of x
        # eg. minxmindig =-4, maxxmindig=-1: variable will have at least 1
        # digit after comma, but not more than 4
        self.minxmindig = None
        self.maxxmindig = None
        
        # min/max vmindig: minimum digit to be used for each var (order of magnitude)
        # can be set by the caller to restrict the precision
        # eg. minvmindig =-4, maxvmindig=-1: variable will have at least 1
        # digit after comma, but not more than 4
        self.minvmindig = None
        self.maxvmindig = None
        
        # self.minvmiss: caller can request to generate a minimal VMISS value
        # for the respective variable (do not add one order of magnitude)
        self.minvmiss = None
        
        # the main data attribute, which represents the file's content
        self.data = None
        self.init_data()
        
        self.logger = logging.getLogger("Xml")

    def init_data(self):
        """
        Initializes the data object. Sets all attributes to None.
        Parameters:
            None
        Returns:
            None
        """
        self.data = DataObject(dict([(att, None)
                                     for att in self.__class__.DATA_ATTRIBS]))

    def init_state(self):
        """
        Initialize the object state before reading or writing operations.
        Parameters:
            None
        Returns:
            None
        """
        self.file = None
        self.state = -1
        self.errors = 0
        self.warnings = 0
        self.lnum = 0

    def curr_state_name(self):
        """
        Returns the name of the current reading state.
        Parameters:
            None
        Returns:
            state name (str)
        """
        return self.__class__.SECTIONS[self.state]

    def dx_from_data(self):
        """
        Calculates DX from the sample intervals.
        Parameters:
            None
        Returns:
            DX (real)
        There are some problematic cases with float representation:
            e.g. the sequence 1.1, 1.2, 1.3, ...
            would result in a dx between 0.09999999999999987
            and 0.10000000000000009
            (even worse with e.g. 1.000000001, 1.000000002, ...)

            Using Decimal instead of float would raise performance issues

            Solution: check dx with float operations (and almostequal)
            Then, in the end calculate a Decimal dx using the first two samples
            only.
        """
        # calculate dx
        if not self.data.DATA:
            return 0
        totaldiff = None
        for diff in zip(self.data.DATA[0], self.data.DATA[0][1:]):
            if diff[0] is None or diff[1] is None:
                raise XmlError(
                    "missing values not allowed for x-values")
            if not totaldiff:
                totaldiff = float(diff[1] - diff[0])
            # if x output precission is restricted by the caller, use this
            # precission as comp. criteria
            if self.xmindig:
                magnitude = self.xmindig
                if self.xmindig >= 1:
                    # xmindig is != order of magnitude
                    magnitude = self.xmindig - 1
                if abs(diff[0] + totaldiff - diff[1]) >= \
                   0.5* 10 ** magnitude:
                    totaldiff = None
                    break
            elif not float_almostequal(float(diff[0] + totaldiff),
                                       float(diff[1])):
                totaldiff = None
                break
        # covers also special case: none ore one sample:
        if totaldiff is None:
            return 0
        return float(Decimal(str(self.data.DATA[0][1])) - \
                     Decimal(str(self.data.DATA[0][0])))

    def error(self, msg, lnum=True, exception=False):
        """
        Error handling for file read.
        Error will be logged, lnum included if passed.
        Parameters:
            msg         error message
            lnum        include line number (optional)
            exception   raises exception if true
        Returns:
            None
        Raises:
            NasaAmesError when parameter exception is True
        """
        self.errors += 1
        if lnum:
            msg = "line {}: {}".format(self.lnum, msg)
        self.logger.error(msg)
        if exception:
            raise XmlError(msg)

    def warning(self, msg, lnum=True):
        """
        Warning message for file read.
        Worning will be logged, lnum included if passed.
        Parameters:
            msg         warning message
            lnum        include line number (optional)
        Returns:
            None
        """
        self.warnings += 1
        if lnum:
            msg = "line {}: {}".format(self.lnum, msg)
        self.logger.warning(msg)




"""
File IO for ASCII datafiles in fixed witdh format.
"""

from nilutility.datatypes import DataObject
import datetime
from six import PY2

class DatetimeParserFactory(object):  # pylint: disable=R0903
    # R0903:Too few public methods
    """
    Factory class for datetime.datetime, constructed by parsing a string.
    """
    def __init__(self, formatstring):
        """
        Set up datetime_parser_factory.
        Parameters:
            None
        Returns:
            None
        """
        self.formatstring = formatstring

    def __call__(self, datetimestr):
        """
        Actually create the object: a datatime.datetime that is parsed from a
        string.
        Parameters:
            datetimestring     string containing the datetime value according to
                               the format string self.formatstr
        Returns:
            datetime.datetime object
        """
        return datetime.datetime.strptime(datetimestr, self.formatstring)

class AsciiFixedwidth(object):
    """
    File IO for ASCII datafiles in fixed witdh format.
    """

    def __init__(self, header=False):
        """
        Set up IO object.

        Parameters:
            None
        Returns:
            None
        Raises:
            None
        """
        self.format_desc = DataObject(header=header, columns=[])
        self.data = []
        self.header = []

    def add_column(self, pos_start, length, datatype=None):
        """
        Adds a column to the format description.

        Parameters:
            pos_start   start position in file
            length      length of the column
            datatype    datatype to be used for the column
        """
        if PY2:
            datatype = unicode
        else:
            datatype = str
        self.format_desc.columns.append(
            DataObject(pos_start=pos_start, length=length,
                       datatype=datatype, header=None))
        self.data.append([])
        self.header.append(None)

    def read(self, filename):
        """
        Reads a file according to the data format description.

        Parameters:
            filename     file path and name

        Returns:
            None
        """
        fil = open(filename)
        linenum = -1
        for line in fil:
            linenum += 1
            if self.format_desc.header and linenum == 0:
                # prosess header
                for col in self.format_desc.columns:
                    col.header = line[col.pos_start:col.pos_start+col.length]\
                                     .strip()
            else:
                # prosess data
                var_index = 0
                for col in self.format_desc.columns:
                    self.data[var_index].append(col.datatype(
                        line[col.pos_start:col.pos_start+col.length].strip()))
                    var_index += 1




"""
$Id: file_helper.py 2614 2021-02-18 10:59:10Z pe $

Module for file io helper functions.
This module is just a helper for dealing with common file  i/o needs.
"""

import os
import re
import hashlib
import six

def tail(file_, nlines, expected_linelength=300):
    """
    Returns the last nline lines from a file.
    Parameters:
        file    opened file handle or file name including path
        nlines  number of lines that should be returned
        expected_linelength
                average expected length of lines for the last n lines
                with files that have very long lines (veeery long, say > 1000),
                this will help to speed up the reading
    Returns:
        The last nlines lines of the file (list of strings).
        Less then n lines, if there is not more in the file.
        The strings are raw data from the file, no decoding, no substitution
        of line end characters!
        AsciiRead.goto_line relies on this!
    Raises:
        IOError:  on various io problems
            file name does not exist, no access, ...)
    """
    if isinstance(file_, six.string_types):
        with open(file_, 'r') as file_:
            return tail(file_, nlines, expected_linelength=expected_linelength)
    if nlines < 0:
        raise ValueError('tail negative line numbers')
    if nlines == 0:
        return []
    # get the file size:
    file_.seek(0, os.SEEK_END)
    filesize = file_.tell()
    chunksize = min(expected_linelength * nlines, filesize)
    lastlines_pattern = r'[^\n\r]*\r?\n?'
    if not six.PY2:
        from io import BytesIO
        is_bin = False
        if isinstance(file_, BytesIO):
            is_bin = True
        else:
            try:
                is_bin = 'b' in file_.mode
            except AttributeError:
                pass
        if is_bin:
            lastlines_pattern = lastlines_pattern.encode('utf-8')
    while True:
        file_.seek(0, os.SEEK_END)
        file_.seek(file_.tell()-chunksize, os.SEEK_SET)
        buf = file_.read()
        lines = re.findall(lastlines_pattern, buf)[0:-1]  # last elem. '', skip
        # NO decoding, NO substitution of line ends: see docstring!
        if len(lines) > nlines:
            # we need nlines+1 to make sure the first line is complete
            # generate result:
            return lines[-nlines:]
        if chunksize == filesize:
            # fine, if we read already the whole file, return what's there
            return lines[-nlines:]
        else:
            # estimate a new buffer size
            if len(lines) <= 1:
                # not even one full line read: double the chunk
                chunksize *= 2
            else:
                # estiomate new size according to the fraction read already
                factor = float(nlines) / float(len(lines)-1)
                chunksize = int(chunksize * factor)
            if chunksize > filesize:
                chunksize = filesize

def md5(file_, chunksize=65536):
    """
    Generates and returns an md5 hash for a file.
    Parameters:
        file_       file object or file path
        chunksize   chunk size for file reading
    Returns:
        md5 hash of the file
    Raises:
       IOError
    """
    def _md5():
        """
        Generate the hash for a file object (opened file)
        Parameters:
            None
        Returns:
            md5 hash
        """
        md5_hash = hashlib.md5()
        for chunk in iter(lambda: file_.read(chunksize), b""):
            md5_hash.update(chunk)
        return md5_hash.hexdigest()

    if isinstance(file_, six.string_types):
        with open(file_, "rb") as file_:
            return _md5()
    else:
        return _md5()

def hardlinktree(src, dst):
    """
    Makes a copy of a directory tree, but hard links the files instead of
    copying.
    Similar to shutil.copytree, but hard links (in python3, there is a
    copy_function parameter, so this function mimics:
       shutil.copytree(src, dst, copy_function=os.link)
    in python3.
    """
    dst = os.path.abspath(dst)
    os.mkdir(dst)
    for root, dirs, files in os.walk(src, topdown=True):
        root = os.path.relpath(root, start=src)
        if root == '.':
            root = ''
        curdst = os.path.join(dst, root)
        for dir_ in dirs:
            os.mkdir(os.path.join(curdst, dir_))
        for file_ in files:
            fromfile = os.path.join(src, root, file_)
            tofile = os.path.join(curdst, file_)
            os.link(fromfile, tofile)

def get_main_path():
    """
    Get the path name of the main script running.
    """
    return os.path.abspath(__main__.__file__)



"""
Data level spcific functionality
"""

def base_datalevel(data_level):
    """
    Returns the base datalevel.
    Parameters:
        datalevel    specific datalevel
    Returns:
        base datalevel (0, 1, 2, 3)
    """
    if data_level in ('0', '0.5', '0a', '0b'):
        return '0'
    if data_level in ('1', '1a', '1b'):
        return '1'
    if data_level in ('1.5', '2', '2a', '2b'):
        return '2'
    if data_level in ('3', '3a', '3b'):
        return '3'

